use super::*;

pub struct CommandDeamon {
    thread: Arc<Mutex<Cell<Option<thread::JoinHandle<()>>>>>,
    quit: Arc<AtomicBool>,
    joined: Arc<AtomicBool>,
}

impl Deamon for CommandDeamon {
    type Channels = (Sender<LektorQuery>, Receiver<LektorResponse>);
    implement_deamon_quit!();
    implement_deamon_joined!();

    fn spawn(hostname: String, port: i16) -> io::Result<(Self::Channels, Self)> {
        let mut connexion = LektorConnexion::new(hostname, port)?;

        let (responses_send, responses_recv) = channel::<LektorResponse>();
        let (commands_send, commands_recv) = channel::<LektorQuery>();
        let quit = Arc::<AtomicBool>::new(AtomicBool::default());
        let joined = Arc::<AtomicBool>::new(AtomicBool::default());
        quit.store(false, atomic::Ordering::SeqCst);
        joined.store(false, atomic::Ordering::SeqCst);
        let quit_deamon = quit.clone();

        let thread = thread::spawn(move || {
            loop {
                break_when_flagged!(quit_deamon);
                match commands_recv.recv_timeout(Duration::from_secs(1)) {
                    Ok(command) => {
                        let res = match connexion.send_query(command) {
                            Ok(ok) => ok,
                            Err(e) => {
                                error!("failed to send query to lektor: {e}");
                                info!("quiting the status deamon");
                                quit_deamon.store(false, atomic::Ordering::SeqCst);
                                break;
                            }
                        };
                        if let Err(e) = responses_send.send(res) {
                            error!("failed to send response to amadeus: {e}");
                            info!("quiting the status deamon");
                            quit_deamon.store(false, atomic::Ordering::SeqCst);
                            break;
                        }
                    }
                    Err(RecvTimeoutError::Timeout) => {}
                    Err(e) => {
                        error!("failed to get command from amadeus: {e}");
                        info!("quiting the command deamon");
                        quit_deamon.store(false, atomic::Ordering::SeqCst);
                        break;
                    }
                };
            }
            info!("quiting the command deamon!");
        });

        let ret = Self {
            thread: Arc::new(Mutex::new(Cell::new(Some(thread)))),
            quit,
            joined,
        };
        Ok(((commands_send, responses_recv), ret))
    }
}
