use std::collections::BTreeMap;

use egui::{
    FontData, FontDefinitions,
    FontFamily::{Monospace, Proportional},
    FontId, TextStyle,
};

#[inline]
pub fn heading1() -> TextStyle {
    TextStyle::Heading
}

#[inline]
pub fn heading2() -> TextStyle {
    TextStyle::Name("Heading2".into())
}

#[inline]
pub fn heading3() -> TextStyle {
    TextStyle::Name("ContextHeading".into())
}

#[inline]
pub fn body() -> TextStyle {
    TextStyle::Body
}

#[inline]
pub fn small_body() -> TextStyle {
    TextStyle::Name("SmallBody".into())
}

#[inline]
pub fn monospace() -> TextStyle {
    TextStyle::Monospace
}

#[inline]
pub fn button() -> TextStyle {
    TextStyle::Button
}

#[inline]
pub fn small() -> TextStyle {
    TextStyle::Small
}

pub fn get_font_definitions() -> FontDefinitions {
    static FONT_NAME: &str = "UbuntuMono";
    static FONT_DATA: &[u8] = include_bytes!("../../../rsc/UbuntuMono-Regular.ttf");
    static CJK_NAME: &str = "IPAMincho";
    static CJK_DATA: &[u8] = include_bytes!("../../../rsc/IPAMincho.ttf");
    let mut fonts = FontDefinitions::default();

    fonts
        .font_data
        .insert(FONT_NAME.to_owned(), FontData::from_static(FONT_DATA));
    fonts
        .font_data
        .insert(CJK_NAME.to_owned(), FontData::from_static(CJK_DATA));

    fonts
        .families
        .get_mut(&Proportional)
        .unwrap()
        .insert(0, FONT_NAME.to_owned());
    fonts
        .families
        .get_mut(&Proportional)
        .unwrap()
        .push(CJK_NAME.to_owned());

    fonts
        .families
        .get_mut(&Monospace)
        .unwrap()
        .insert(0, FONT_NAME.to_owned());
    fonts
        .families
        .get_mut(&Monospace)
        .unwrap()
        .push(CJK_NAME.to_owned());

    fonts
}

pub fn get_font_styles() -> BTreeMap<TextStyle, FontId> {
    let styles = [
        (heading1(), FontId::new(30.0, Proportional)),
        (heading2(), FontId::new(25.0, Proportional)),
        (heading3(), FontId::new(23.0, Proportional)),
        (body(), FontId::new(18.0, Proportional)),
        (monospace(), FontId::new(14.0, Proportional)),
        (small_body(), FontId::new(14.0, Proportional)),
        (button(), FontId::new(14.0, Proportional)),
        (small(), FontId::new(10.0, Proportional)),
    ];
    styles.into()
}
