mod about_window;
mod action_menu;
mod playlists;
mod progress_bar;
mod toggle_switch;
mod window;

pub use about_window::render_about_window;
pub use action_menu::render_action_menu;
use lkt_lib::Kara;
pub use playlists::PlaylistsStore;
pub use progress_bar::progress_bar;
pub use toggle_switch::toggle_switch;
pub use window::WindowBuilder;

use crate::{constants, utils};

pub fn add_bullet_hyperlink(ui: &mut egui::Ui, label: &str, url: &str, after: Option<&str>) {
    ui.horizontal(|ui| {
        ui.label("    -");
        ui.hyperlink_to(label, url);
        if let Some(text) = after {
            ui.add(egui::Label::new(text).wrap(true));
        }
    });
}

pub fn add_heading2_label(ui: &mut egui::Ui, text: &str) {
    ui.style_mut().override_text_style = Some(utils::font::heading2());
    ui.add(egui::Label::new(text).wrap(false));
    ui.add_space(constants::PADDING);
}

pub fn add_labelled_edit_line(
    ui: &mut egui::Ui,
    before: &str,
    text: &mut dyn egui::TextBuffer,
    changed: &mut bool,
) {
    ui.horizontal(|ui| {
        ui.style_mut().override_text_style = Some(utils::font::body());
        ui.add(egui::Label::new(before));
        let response = ui.add(egui::TextEdit::singleline(text));
        *changed |= response.lost_focus();
    });
}

pub fn add_labelled_toggle_switch(
    ui: &mut egui::Ui,
    dark_mode: bool,
    before: &str,
    switch: &mut bool,
    changed: &mut bool,
) {
    ui.horizontal(|ui| {
        ui.style_mut().override_text_style = Some(utils::font::body());
        ui.add(egui::Label::new(before));
        ui.with_layout(egui::Layout::right_to_left(egui::Align::Min), |ui| {
            let response = ui.add(toggle_switch(dark_mode, switch));
            *changed |= response.changed();
        });
    });
}

pub fn add_labelled_password(
    ui: &mut egui::Ui,
    before: &str,
    text: &mut dyn egui::TextBuffer,
    changed: &mut bool,
) {
    ui.horizontal(|ui| {
        ui.style_mut().override_text_style = Some(utils::font::body());
        ui.add(egui::Label::new(before));
        let response = ui.add(egui::TextEdit::singleline(text).password(true));
        *changed |= response.lost_focus();
    });
}

pub fn get_label_text_for_kara(kara: &Kara) -> String {
    format!(
        "{} - {} / {} - {} - {} [{}]",
        kara.category, kara.language, kara.source_name, kara.song_type, kara.title, kara.author
    )
}
